/*
 * wlanmodel/packetproxy.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "PacketProxy.h"


/** Register as simulation module */
Define_Module(PacketProxy);

/**
 * Constructor
 */
PacketProxy::PacketProxy()
{
    rtEvent = NULL;
}

/**
 * Destructor
 */
PacketProxy::~PacketProxy()
{
    cancelAndDelete(rtEvent);
}

/**
 * Initialize PacketProxy at simulation startup
 */
void
PacketProxy::initialize()
{
    rtEvent = new cMessage("rtEvent");
    rtScheduler =
            check_and_cast<cUDPSocketRTScheduler *>(simulation.getScheduler());
    rtScheduler->setInterfaceModule(this, rtEvent, inBuffer,
                                    MAX_MSG_SIZE, &numRecvBytes);
    nodeManager =
        check_and_cast<NodeManager *>(simulation.getModuleByPath("nodemanager"));
    pktProcTime.setName("Packet Processing Time");
}

/**
 * PacketProxy is only notified when the scheduler receives a UDP packet.
 * Check the buffer for the message type then act accordingly. If unknown
 * message, discard it.
 */
void
PacketProxy::handleMessage(cMessage *msg)
{
    //	debugMessage(inBuffer, numRecvBytes);
	switch(getMessageType((msgHeader *) inBuffer))
	{
		case DATA_MSG:
			ev << "PacketProxy.handleSocketEvent(): WLAN_MSG received\n";
            printDataMessage(inBuffer);
            processDataMsg((dataMsg *) inBuffer);
            break;
		case REGISTRATION_MSG:
            ev << "PacketProxy.handleSocketEvent(): REGISTRATION_MSG received\n";
            printRegisterMessage(inBuffer);
            processRegistrationMsg((registrationMsg *) inBuffer);
            break;
		case DEREGISTRATION_MSG:
            ev << "PacketProxy.handleSocketEvent(): DEREGISTRATION_MSG received\n";
            printPlainMessage(inBuffer);
            processDeregistrationMsg((plainMsg *) inBuffer);
            break;
        case CONFIG_MSG:
            ev << "PacketProxy.handleSocketEvent(): CONFIG_MSG received\n";
        	printConfigMessage(inBuffer);
        	processConfigurationMsg((configMsg *) inBuffer);
            break;
        default:
            ev << "PacketProxy.handleSocketEvent(): Unknown message received ("
               << getMessageType((msgHeader *) inBuffer) << ")\n";
            numRecvBytes = 0;
            break;
    }
}

/**
 * If host is registered send wlanPkt via host's MsgHandler to model
 */
void
PacketProxy::processDataMsg(dataMsg *msg)
{
    host_id *hostId = (host_id *) &((msg->header).id);
    // don't use check_and_cast because pointer may be NULL
    cModule *host = (cModule *) nodeManager->isRegistered(hostId);

    if (host != NULL) // if no such host, just discard message
    {
    	MsgHandler *handler =
                check_and_cast<MsgHandler *>(host->getSubmodule("msgHandler"));
    	handler->sendToModel((char *) msg->wlanPkt, msg->pktLength);
    }

    // clean input buffer
    numRecvBytes = 0;
}

/**
 * If host is registered acknowlege registration, else register host
 */
void
PacketProxy::processRegistrationMsg(registrationMsg *msg)
{
    host_id *hostId = (host_id *) &((msg->header).id);
    // don't use check_and_cast because pointer may be NULL
    cModule *host = (cModule *) nodeManager->isRegistered(hostId);

    if (host == NULL)
    {
    	nodeManager->addNode(msg);
    } else {
    	unsigned int msgId = (msg->info).msgId;
    	MsgHandler *handler =
                check_and_cast<MsgHandler *>(host->getSubmodule("msgHandler"));
    	handler->sendAck(msgId);
    }

    // clean input buffer
    numRecvBytes = 0;
}

/**
 * If host is registered remove host modul and ack
 */
void
PacketProxy::processDeregistrationMsg(plainMsg *msg)
{
    host_id *hostId = (host_id *) &((msg->header).id);
    // don't use check_and_cast because pointer may be NULL
    cModule *host = (cModule *) nodeManager->isRegistered(hostId);

    if (host != NULL)
    	nodeManager->removeNode(host);

    // clean input buffer
    numRecvBytes = 0;
}

/**
 * Dispose wireless device parameter change
 */
void
PacketProxy::processConfigurationMsg(configMsg *msg)
{
    host_id *hostId = (host_id *) &((msg->header).id);
    // don't use check_and_cast because pointer may be NULL
    cModule *host = (cModule *) nodeManager->isRegistered(hostId);

    if (host != NULL) // if no such host, just discard message
    {
    	MsgHandler *handler =
                check_and_cast<MsgHandler *>(host->getSubmodule("msgHandler"));
    	handler->updateVifParameter(&(msg->vifData));
    }

    // clean input buffer
    numRecvBytes = 0;
}

/**
 * Return message type
 */
unsigned char
PacketProxy::getMessageType(msgHeader *header)
{
    return header->type;
}

/**
 * Print message
 */
void
PacketProxy::debugMessage(const char *msg, int length)
{
    ev.printf("\n(%d bytes) data frame:\n", length);
    for (int i = 0; i < length; i+=2)
    {
    	if (i % 16 == 0)
    	{
    		if (i == 0)
    			ev.printf("(0x0000) ");
    		else
    			ev.printf("(%#.4x) ", i);
    	}
        ev.printf("%02hhx%02hhx ", *(msg + i), *(msg + (i + 1)));
        if (((i + 2) % 16 == 0) && (i > 0))
			ev.printf("\n");
    }
    ev.printf("\n");
}

void
PacketProxy::printConfigMessage(const char *message)
{
    msgHeader *header = (msgHeader *) message;
    ev.printf("\n[%4lu bytes] ========== CONFIGURE MESSAGE ===========\n",
            CONFIG_MSG_SIZE);
    ev.printf("[%4lu bytes] ,---------- msgHeader -------------------\n",
            sizeof(msgHeader));
    ev.printf("(%4lu bytes) | type:\t%i\n", sizeof(header->type), header->type);
    ev.printf("(%4lu bytes) | id:\t%s\n", sizeof(header->id), header->id);

	configNotification info =
                        (configNotification) ((configMsg *) message)->vifData;
    uint16_t type = info.type;
    vifParameter *value_p = &(info.value);
    ev.printf("(%4lu bytes) ,---------- configNotification ----------\n",
            sizeof(configNotification));
    switch (type)
    {
        case SIOCSIWFREQ:
            ev.printf("(%4lu bytes) | property:\t%4x (SIOCSIWFREQ)\n",
                    sizeof(type), type);
            ev.printf("(%4lu bytes) | value:\t\t%.3e\n",
                    sizeof(*value_p), value_p->freq);
            break;
        case SIOCSIWTXPOW:
        	ev.printf("(%4i bytes) | property:\t%4x (SICSIWPOW)\n",
                    sizeof(type), type);
        	ev.printf("(%4i bytes) | value:\t\t%imW\n",
                    sizeof(*value_p), value_p->txpower);
            break;
        case SIOCSIWRETRY:
        	ev.printf("(%4i bytes) | property:\t%4x (SICSIWRETRY)\n",
                    sizeof(type), type);
        	ev.printf("(%4i bytes) | value:\t\t%i\n",
                    sizeof(*value_p), value_p->retry);
            break;
        case SIOCSIWRTS:
        	ev.printf("(%4i bytes) | property:\t%4x (SICSIWRTS)\n",
                    sizeof(type), type);
        	ev.printf("(%4i bytes) | value:\t\t%iB\n",
                    sizeof(*value_p), value_p->rts);
            break;
        case SIOCSIWSENS:
        	ev.printf("(%4i bytes) | property:\t%4x (SICSIWSENS)\n",
                    sizeof(type), type);
        	ev.printf("(%4i bytes) | value:\t\t%idB\n",
                    sizeof(*value_p), value_p->sensitivity);
            break;
        default:
        	ev.printf("(%4i bytes) | property:\t\t%4x (unknown)\n",
                    sizeof(type), type);
            break;
    }
    printf("------------------------------------------------------\n");
}

void
PacketProxy::printRegisterMessage(const char *msg)
{
    msgHeader *header = (msgHeader *) msg;
    ev.printf("\n[%4lu bytes] ========== REGISTRATION MESSAGE ========\n",
            sizeof(registrationMsg));
    ev.printf("[%4lu bytes] ,---------- msgHeader -------------------\n",
            sizeof(msgHeader));
    ev.printf("(%4lu bytes) | type:\t%i\n", sizeof(header->type), header->type);
    ev.printf("(%4lu bytes) | id:\t%s\n", sizeof(header->id), header->id);

    registrationInfo *regInfo = &((registrationMsg *) msg)->info;
    ev.printf("[%4lu bytes] ,---------- registrationInfo ------------\n",
            sizeof(registrationInfo));
    ev.printf("(%4lu bytes) | name:\t%s\n", sizeof(regInfo->name), regInfo->name);
    ev.printf("[%4lu bytes] | ,---------------- addr ----------------\n",
            sizeof(regInfo->addr));
    ev.printf("(%4lu bytes) | | sin_family:\t%i\n",
            sizeof(regInfo->addr.ss_family), regInfo->addr.ss_family);
    if (regInfo->addr.ss_family == AF_INET)
    {
    	ev.printf("(%4i bytes) | | sin_addr:\t%s\n", sizeof(struct in_addr),
            inet_ntoa(((struct sockaddr_in *) &regInfo->addr)->sin_addr));
    	ev.printf("(%4i bytes) | | sin_port:\t%i\n", sizeof(unsigned short),
            ntohs(((struct sockaddr_in *) &regInfo->addr)->sin_port));
    }
    if (regInfo->addr.ss_family == AF_INET6)
    {
        char addr_string[INET6_ADDRSTRLEN];
        ev.printf("(%4i bytes) | | sin6_addr:\t%s\n", sizeof(struct in6_addr),
            inet_ntop(AF_INET6,
                    &((struct sockaddr_in6 *) &regInfo->addr)->sin6_addr,
                    addr_string, INET6_ADDRSTRLEN));
        ev.printf("(%4i bytes) | | sin6_port:\t%i\n", sizeof(uint16_t),
            ntohs(((struct sockaddr_in6 *) &regInfo->addr)->sin6_port));
    }
    ev.printf("(%4lu bytes) | msgId:\t%i\n",
            sizeof(regInfo->msgId), regInfo->msgId);
    ev.printf("(%4lu bytes) | numVif:\t%i\n",
            sizeof(regInfo->numVif), regInfo->numVif);

    for (int i = 0; i < regInfo->numVif; i++)
    {
    	vifRegistrationData *vifInfo =
                                (vifRegistrationData *) &regInfo->vifData[i];
        ev.printf("[%4lu bytes] | ,---------------- vifRegistrationData -\n",
                sizeof(vifRegistrationData));
        ev.printf("(%4i bytes) | | vif_name: \t%s\n",
                IFNAMSIZ + 1, vifInfo->vifName);
        unsigned char *mac = vifInfo->vifMac;
        ev.printf("(%4i bytes) | | vif_mac:\t%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
                ETH_ALEN, ((unsigned char *) mac)[0],
                ((unsigned char *) mac)[1], ((unsigned char *) mac)[2],
                ((unsigned char *) mac)[3], ((unsigned char *) mac)[4],
                ((unsigned char *) mac)[5]);
        ev.printf("(%4i bytes) | | vifIndex:\t%i\n",
                sizeof(int16_t), vifInfo->vifIndex);
        ev.printf("(%4i bytes) | | txpower:\t%imW\n",
                sizeof(int16_t), vifInfo->txpower);
        ev.printf("(%4i bytes) | | retry:\t\t%i\n",
                sizeof(int16_t), vifInfo->retry);
        ev.printf("(%4i bytes) | | rts:\t\t%iB\n",
                sizeof(int16_t), vifInfo->rts);
        ev.printf("(%4i bytes) | | sensitivity:\t%idB\n",
                sizeof(int16_t), vifInfo->sensitivity);
        ev.printf("(%4i bytes) | | freq:\t\t%.3e\n",
                sizeof(int64_t), vifInfo->freq);
    }
}

void
PacketProxy::printPlainMessage(const char *message)
{
    msgHeader *header = (msgHeader *) message;
    if (header->type == ACK_MSG)
        ev.printf("\n[%4lu bytes] ========== ACKNOWLEDGE MESSAGE =========\n",
                sizeof(plainMsg));
    if (header->type == DEREGISTRATION_MSG)
        ev.printf("\n[%4lu bytes] ========== DEREGISTRATION MESSAGE ======\n",
                sizeof(plainMsg));
    ev.printf("[%4lu bytes] ,---------- msgHeader -------------------\n",
            sizeof(msgHeader));
    ev.printf("(%4lu bytes) | type:\t%i\n", sizeof(header->type), header->type);
    ev.printf("(%4lu bytes) | id:\t%s\n", sizeof(header->id), header->id);

    plainMsg *msg = (plainMsg *) message;
    ev.printf("(%4lu bytes) msg_id:\t%i\n", sizeof(uint8_t), msg->msgId);
}

void
PacketProxy::printDataMessage(const char *message)
{
	msgHeader *header = (msgHeader *) message;
	dataMsg *msg = (dataMsg *) message;
	int payloadSize = msg->pktLength;
    ev.printf("\n[%4lu bytes] ----------- DATA MESSAGE ----------------\n",
            DATA_PREAMBLE_SIZE + payloadSize);
    ev.printf("[%4lu bytes] ,---------- msgHeader -------------------\n",
            sizeof(msgHeader));
    ev.printf("(%4lu bytes) | type:\t%i\n", sizeof(header->type), header->type);
    ev.printf("(%4lu bytes) | id:\t%s\n", sizeof(header->id), header->id);
    ev.printf("(%4lu bytes) vifIndex:\t%u\n",
            sizeof(msg->vifIndex), msg->vifIndex);
    ev.printf("(%4lu bytes) pktLength:\t%u\n",
            sizeof(msg->pktLength), payloadSize);
    debugMessage((char *) msg->wlanPkt, payloadSize);
}
