/*
 * wlanmodel/cUDPSocketRTScheduler.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <netdb.h>
#include "cUDPSocketRTScheduler.h"

Register_Class(cUDPSocketRTScheduler);
Register_GlobalConfigOption(CFGID_UDPSOCKETRTSCHEDULER_PORT, "udpsocketrtscheduler-port", CFG_INT, "2424", "When cUDPSocketRTScheduler is selected as scheduler class: the port number cUDPSocketRTScheduler listens on.");

/**
 * Constructor
 */
cUDPSocketRTScheduler::cUDPSocketRTScheduler() : cScheduler()
{
    listenerSocket = INVALID_SOCKET;
    connSocket = INVALID_SOCKET;
}

/**
 * Destructor
 */
cUDPSocketRTScheduler::~cUDPSocketRTScheduler()
{
}

/**
 * Called at the beginning of a simulation run.
 */
void
cUDPSocketRTScheduler::startRun()
{
    if (initsocketlibonce()!=0)
        throw new cRuntimeError("cUDPSocketRTScheduler: Cannot initialize socket library");

    gettimeofday(&baseTime, NULL);

    module = NULL;
    notificationMsg = NULL;
    recvBuffer = NULL;
    recvBufferSize = 0;
    numBytesPtr = NULL;

    port = ev.getConfig()->getAsInt(CFGID_UDPSOCKETRTSCHEDULER_PORT);
    setupListener();
}

/**
 * Called at the end of a simulation run.
 */
void
cUDPSocketRTScheduler::endRun()
{
}

/**
 * Setup listening socket for receiving node traffic
 */
void
cUDPSocketRTScheduler::setupListener()
{
	struct addrinfo request;
	struct addrinfo *result;

	memset(&request, 0, sizeof(struct addrinfo));
	request.ai_family = AF_UNSPEC;		// allow IPv4 and IPv6
	request.ai_socktype = SOCK_DGRAM;	// UDP socket
	request.ai_flags = AI_PASSIVE;		// for wildcard IP address
	request.ai_protocol = 0;			// any protocol
	request.ai_canonname = NULL;
	request.ai_addr = NULL;
	request.ai_next = NULL;

	char cport[5];						// max 65536
	sprintf(cport, "%i", port);

	if (getaddrinfo(NULL, cport, &request, &result) < 0)
		throw new cRuntimeError("cUDPSocketRTScheduler: getaddrinfo");

    listenerSocket = socket(result->ai_family, result->ai_socktype, result->ai_protocol);
    if (listenerSocket == INVALID_SOCKET)
        throw new cRuntimeError("cUDPSocketRTScheduler: cannot create socket");

    if (bind(listenerSocket, result->ai_addr, result->ai_addrlen) == SOCKET_ERROR)
        throw new cRuntimeError("cUDPSocketRTScheduler: socket bind() failed");
}

/**
 * Called every time the user hits the Run button in Tkenv.
 * Use this callback to pin current simulation time to current wall clock time.
 */
void
cUDPSocketRTScheduler::executionResumed()
{
    gettimeofday(&baseTime, NULL);
    baseTime = timeval_substract(baseTime, sim->getSimTime().dbl());
}

void
cUDPSocketRTScheduler::setInterfaceModule(cModule *mod, cMessage *notifMsg, \
    char *buf, int bufSize, int *nBytesPtr)
{
    if (module)
        throw new cRuntimeError("cUDPSocketRTScheduler: setInterfaceModule() already called");
    if (!mod || !notifMsg || !buf || !bufSize || !nBytesPtr)
        throw new cRuntimeError("cUDPSocketRTScheduler: setInterfaceModule(): arguments must be non-NULL");

    module = mod;
    notificationMsg = notifMsg;
    recvBuffer = buf;
    recvBufferSize = bufSize;
    numBytesPtr = nBytesPtr;
    *numBytesPtr = 0;
}

/**
 * Listen for incoming network traffic until timeout exceeds
 */
bool cUDPSocketRTScheduler::receiveWithTimeout(long usec)
{
    // prepare sets for select()
    fd_set readFDs, writeFDs, exceptFDs;
    FD_ZERO(&readFDs);
    FD_ZERO(&writeFDs);
    FD_ZERO(&exceptFDs);

    FD_SET(listenerSocket, &readFDs);

    timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (select(FD_SETSIZE, &readFDs, &writeFDs, &exceptFDs, &timeout) > 0)
    {
        if (listenerSocket != INVALID_SOCKET && FD_ISSET(listenerSocket, &readFDs))
        {
            // receive from listenerSocket
            sockaddr_in sinRemote;
            int addrSize = sizeof(sinRemote);
            char *bufPtr = recvBuffer + (*numBytesPtr);
            int bufLeft = recvBufferSize - (*numBytesPtr);
            if (bufLeft <= 0)
                throw new cRuntimeError("cUDPSocketRTScheduler: interface module's recvBuffer is full");
            int nBytes = recvfrom(listenerSocket, bufPtr, bufLeft, 0, \
                (sockaddr*) &sinRemote, (socklen_t*) &addrSize);
            if (nBytes == SOCKET_ERROR)
            {
                ev << "cUDPSocketRTScheduler: error receiving socket" << sock_errno() << "\n";
                connSocket = INVALID_SOCKET;
            } else {
                // schedule notificationMsg for the interface module
                ev << "cUDPSocketRTScheduler: " << nBytes << " bytes received\n";
                (*numBytesPtr) += nBytes;

                notificationMsg->setName("rtmsg");

                timeval curTime;
                gettimeofday(&curTime, NULL);
                curTime = timeval_substract(curTime, baseTime);
                simtime_t t = curTime.tv_sec + curTime.tv_usec * 1e-6;
                // TBD assert that it's somehow not smaller than previous event's time
                notificationMsg->setArrival(module,-1,t);
                notificationMsg->setTimestamp(t);

                simulation.msgQueue.insert(notificationMsg);
                return true;
            }
        }
    }
    return false;
}

int
cUDPSocketRTScheduler::receiveUntil(const timeval& targetTime)
{
    // if there's more than 200ms to wait, wait in 100ms chunks
    // in order to keep UI responsiveness by invoking ev.idle()
    timeval curTime;
    gettimeofday(&curTime, NULL);
    while (targetTime.tv_sec-curTime.tv_sec >=2 ||
           timeval_diff_usec(targetTime, curTime) >= 200000)
    {
        if (receiveWithTimeout(100000)) // 100ms
            return 1;
        if (ev.idle())
            return -1;
        gettimeofday(&curTime, NULL);
    }

    // difference is now at most 100ms, do it at once
    long usec = timeval_diff_usec(targetTime, curTime);
    if (usec>0)
        if (receiveWithTimeout(usec))
            return 1;
    return 0;
}

/**
 * The scheduler function -- it should return the next event to be processed.
 * Normally (with sequential execution) it just returns msgQueue.peekFirst().
 * With parallel and/or real-time simulation, it is also the scheduler's task
 * to synchronize with real time and/or with other partitions.
 */
cMessage *cUDPSocketRTScheduler::getNextEvent()
{
    ev << "Processing next event\n";

    // assert that we've been configured
    if (!module)
        throw new cRuntimeError("cUDPSocketRTScheduler: setInterfaceModule() \
not called: it must be called from a module's initialize() function");

    // calculate target time
    timeval targetTime;
    cMessage *msg = sim->msgQueue.peekFirst();
    if (!msg)
    {
        // if there are no events, wait until something comes from outside
        // TBD: obey simtimelimit, cpu-time-limit
        targetTime.tv_sec = LONG_MAX;
        targetTime.tv_usec = 0;
    }
    else
    {
        // use time of next event
        simtime_t eventSimtime = msg->getArrivalTime();
        targetTime = timeval_add(baseTime, eventSimtime.dbl());
    }

    // if needed, wait until that time arrives
    timeval curTime;
    gettimeofday(&curTime, NULL);
    if (timeval_greater(targetTime, curTime))
    {
        int status = receiveUntil(targetTime);
        if (status == -1)
            return NULL; // interrupted by user
        if (status == 1)
            msg = sim->msgQueue.peekFirst(); // received something
    }
    else
    {
        // we're behind -- customized versions of this class may
        // alert if we're too much behind, whatever that means
    }

    // ok, return the message
    return msg;
}

int
cUDPSocketRTScheduler::sendMessage(const char *msg,
								   const size_t numBytes,
								   const struct sockaddr *nodeAddr,
								   const socklen_t addrSize)
{
    int bytesSent = sendto(listenerSocket, msg, numBytes, 0, nodeAddr, addrSize);
    ev << "cUDPSocketRTScheduler: " << bytesSent << " bytes sent to ";
//    ev.printf("%s:%i", inet_ntoa(node_addr->sin_addr), ntohs(node_addr->sin_port));
    if (bytesSent <= 0)
    {
    	ev << "(" << strerror(errno) << ")\n";
        return SOCKET_ERROR;
    } else {
    	ev << "\n";
        return 0;
    }
}
